# ==============================================================================
#  Copyright (c) 2024 Federico Busetti                                         =
#  <729029+febus982@users.noreply.github.com>                                  =
#                                                                              =
#  Permission is hereby granted, free of charge, to any person obtaining a     =
#  copy of this software and associated documentation files (the "Software"),  =
#  to deal in the Software without restriction, including without limitation   =
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    =
#  and/or sell copies of the Software, and to permit persons to whom the       =
#  Software is furnished to do so, subject to the following conditions:        =
#                                                                              =
#  The above copyright notice and this permission notice shall be included in  =
#  all copies or substantial portions of the Software.                         =
#                                                                              =
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  =
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    =
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     =
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  =
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     =
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         =
#  DEALINGS IN THE SOFTWARE.                                                   =
# ==============================================================================

__version__ = "0.0.3"
__version_tuple__ = (0, 0, 3)



# ==============================================================================
#  Copyright (c) 2024 Federico Busetti                                         =
#  <729029+febus982@users.noreply.github.com>                                  =
#                                                                              =
#  Permission is hereby granted, free of charge, to any person obtaining a     =
#  copy of this software and associated documentation files (the "Software"),  =
#  to deal in the Software without restriction, including without limitation   =
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    =
#  and/or sell copies of the Software, and to permit persons to whom the       =
#  Software is furnished to do so, subject to the following conditions:        =
#                                                                              =
#  The above copyright notice and this permission notice shall be included in  =
#  all copies or substantial portions of the Software.                         =
#                                                                              =
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  =
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    =
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     =
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  =
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     =
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         =
#  DEALINGS IN THE SOFTWARE.                                                   =
# ==============================================================================
from typing import List, TypeVar

from pydantic import TypeAdapter

from ..events import CloudEvent

_T = TypeVar("_T", bound=CloudEvent)


def to_json(event: CloudEvent) -> str:
    """
    Serializes an event in JSON format.

    :param event: The event object to serialize
    :type event: CloudEvent
    :return: The headers and the body representation of the event
    :rtype: str
    """
    return event.model_dump_json()


def from_json(
    data: str, event_adapter: TypeAdapter[_T] = TypeAdapter(CloudEvent)
) -> _T:
    """
    Deserializes an event from JSON format.

    :param data: the JSON representation of the event
    :type data: str
    :param event_adapter: The event class to build
    :type event_adapter: Type[CloudEvent]
    :return: The deserialized event
    :rtype: CloudEvent
    """
    return event_adapter.validate_json(data)


def to_json_batch(
    events: List[_T],
    batch_adapter: TypeAdapter[List[_T]] = TypeAdapter(List[CloudEvent]),
) -> str:
    """
    Serializes a list of events in JSON batch format.

    :param events: The event object to serialize
    :type events: List[CloudEvent]
    :param batch_adapter: The pydantic TypeAdapter to use
    :type: TypeAdapter[List[CloudEvent]]
    :return: The serialized event batch
    :rtype: str
    """
    return batch_adapter.dump_json(events).decode()


def from_json_batch(
    data: str,
    batch_adapter: TypeAdapter[List[_T]] = TypeAdapter(List[CloudEvent]),
) -> List[_T]:
    """
    Deserializes a list of events from JSON batch format.

    :param data: The JSON representation of the event batch
    :type data: str
    :param batch_adapter: The pydantic TypeAdapter to use
    :type: TypeAdapter[List[CloudEvent]]
    :return: The deserialized event batch
    :rtype: List[CloudEvent]
    """
    return batch_adapter.validate_json(data)



# ==============================================================================
#  Copyright (c) 2024 Federico Busetti                                         =
#  <729029+febus982@users.noreply.github.com>                                  =
#                                                                              =
#  Permission is hereby granted, free of charge, to any person obtaining a     =
#  copy of this software and associated documentation files (the "Software"),  =
#  to deal in the Software without restriction, including without limitation   =
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    =
#  and/or sell copies of the Software, and to permit persons to whom the       =
#  Software is furnished to do so, subject to the following conditions:        =
#                                                                              =
#  The above copyright notice and this permission notice shall be included in  =
#  all copies or substantial portions of the Software.                         =
#                                                                              =
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  =
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    =
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     =
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  =
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     =
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         =
#  DEALINGS IN THE SOFTWARE.                                                   =
# ==============================================================================
from typing import (
    Dict,
    Generic,
    List,
    NamedTuple,
    Type,
    TypeVar,
    cast,
)

from pydantic import TypeAdapter

from cloudevents_pydantic.events import CloudEvent
from cloudevents_pydantic.formats import json

_T = TypeVar("_T", bound=CloudEvent)


class HTTPComponents(NamedTuple):
    headers: Dict[str, str]
    body: str


class HTTPHandler(Generic[_T]):
    event_adapter: TypeAdapter[_T]
    batch_adapter: TypeAdapter[List[_T]]

    def __init__(self, event_class: Type[_T] = cast(Type[_T], CloudEvent)) -> None:
        super().__init__()
        self.event_adapter = TypeAdapter(event_class)
        self.batch_adapter = TypeAdapter(List[event_class])  # type: ignore[valid-type]

    def to_json(self, event: _T) -> HTTPComponents:
        """
        Serializes an event in JSON format.

        :param event: The event object to serialize
        :type event: CloudEvent
        :return: The headers and the body representation of the event
        :rtype: HTTPComponents
        """
        headers = {"content-type": "application/cloudevents+json; charset=UTF-8"}
        body = json.to_json(event)
        return HTTPComponents(headers, body)

    def to_json_batch(self, events: List[_T]) -> HTTPComponents:
        """
        Serializes a list of events in JSON batch format.

        :param events: The event object to serialize
        :type events: List[CloudEvent]
        :return: The headers and the body representation of the event batch
        :rtype: HTTPComponents
        """
        headers = {"content-type": "application/cloudevents-batch+json; charset=UTF-8"}
        body = json.to_json_batch(events, self.batch_adapter)
        return HTTPComponents(headers, body)

    def from_json(
        self,
        body: str,
    ) -> CloudEvent:
        """
        Deserializes an event from JSON format.

        :param body: The JSON representation of the event
        :type body: str
        :return: The deserialized event
        :rtype: CloudEvent
        """
        return json.from_json(body, self.event_adapter)

    def from_json_batch(
        self,
        body: str,
    ) -> List[_T]:
        """
        Deserializes a list of events from JSON batch format.

        :param body: The JSON representation of the event batch
        :type body: str
        :return: The deserialized event batch
        :rtype: List[CloudEvent]
        """
        return json.from_json_batch(body, self.batch_adapter)



# ==============================================================================
#  Copyright (c) 2024 Federico Busetti                                         =
#  <729029+febus982@users.noreply.github.com>                                  =
#                                                                              =
#  Permission is hereby granted, free of charge, to any person obtaining a     =
#  copy of this software and associated documentation files (the "Software"),  =
#  to deal in the Software without restriction, including without limitation   =
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    =
#  and/or sell copies of the Software, and to permit persons to whom the       =
#  Software is furnished to do so, subject to the following conditions:        =
#                                                                              =
#  The above copyright notice and this permission notice shall be included in  =
#  all copies or substantial portions of the Software.                         =
#                                                                              =
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  =
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    =
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     =
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  =
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     =
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         =
#  DEALINGS IN THE SOFTWARE.                                                   =
# ==============================================================================
import base64
from datetime import datetime
from enum import Enum
from typing import Annotated, Union
from urllib.parse import ParseResult, urlparse, urlunparse

from annotated_types import Ge, Le
from pydantic import (
    PlainSerializer,
    PlainValidator,
    StringConstraints,
)


def bool_serializer(value: bool) -> str:
    return str(value).lower()


def binary_serializer(value: bytes) -> str:
    return base64.b64encode(value).decode()


def binary_validator(value: Union[str, bytes, bytearray, memoryview]) -> bytes:
    if isinstance(value, (bytes, bytearray, memoryview)):
        return value

    if isinstance(value, str):
        return base64.b64decode(value, validate=True)

    raise ValueError(f"Unsupported value type: {type(value)} - {value}")


def url_serializer(value: ParseResult) -> str:
    return urlunparse(value)


def absolute_uri_validator(value: str) -> ParseResult:
    url = generic_uri_validator(value)
    if not url.scheme:
        raise ValueError("Can't validate absolute URI without scheme")

    return url


def generic_uri_validator(value: str) -> ParseResult:
    if value is None:
        raise ValueError("Field is required")

    return urlparse(value)


"""
Prevents values specified on: https://github.com/cloudevents/spec/blob/v1.0.2/cloudevents/spec.md#type-system
- U+0000-U+001F and U+007F-U+009F ranges - control characters
- non characters specified at: https://www.unicode.org/faq/private_use.html#noncharacters
    - U+FDD0-U+FDEF range
    - U+FFFE and U+FFFF
    - U+1FFFE, U+1FFFF, U+2FFFE, U+2FFFF, ... U+10FFFE, U+10FFFF

We should also check for U+D800-U+DBFF and U+DC00-U+DFFF unless used in pair
(malformed surrogate characters) but pydantic is unhappy to handle them anyway
so we can avoid the scenario in the regex and make this faster.
"""
class_control = r"\u0000-\u001F\u007F-\u009F"
class_nonchar_utf16_range = r"\uFDD0-\uFDEF\uFFFE\uFFFF"
class_nonchar_utf32_range = (
    r"\u{1FFFE}\u{1FFFF}"
    r"\u{2FFFE}\u{2FFFF}"
    r"\u{3FFFE}\u{3FFFF}"
    r"\u{4FFFE}\u{4FFFF}"
    r"\u{5FFFE}\u{5FFFF}"
    r"\u{6FFFE}\u{6FFFF}"
    r"\u{7FFFE}\u{7FFFF}"
    r"\u{8FFFE}\u{8FFFF}"
    r"\u{9FFFE}\u{9FFFF}"
    r"\u{AFFFE}\u{AFFFF}"
    r"\u{BFFFE}\u{BFFFF}"
    r"\u{CFFFE}\u{CFFFF}"
    r"\u{DFFFE}\u{DFFFF}"
    r"\u{EFFFE}\u{EFFFF}"
    r"\u{FFFFE}\u{FFFFF}"
    r"\u{10FFFE}\u{10FFFF}"
)
str_constraint = (
    r"^"
    r"[^"
    + class_control
    + class_nonchar_utf16_range
    + class_nonchar_utf32_range
    + r"]+"
    r"$"
)

str_constraint_asyncapi_compat = (
    r"^" r"[^" + class_control + class_nonchar_utf16_range + r"]+" r"$"
)


# TODO: Add types docstrings
Boolean = Annotated[bool, PlainSerializer(bool_serializer)]
"""
A boolean value of "true" or "false"
"""

Integer = Annotated[int, Ge(-2147483648), Le(2147483648)]
"""
A whole number in the range -2,147,483,648 to +2,147,483,647 inclusive
"""

String = Annotated[str, StringConstraints(pattern=str_constraint_asyncapi_compat)]
"""
Sequence of allowable Unicode characters
"""

Binary = Annotated[
    bytes,
    PlainValidator(binary_validator),
    PlainSerializer(binary_serializer),
]
"""
Sequence of bytes that accepts both bytes and base64 encoded strings as input
and is serialized to a base64 encoded string.
"""

URI = Annotated[
    ParseResult, PlainValidator(absolute_uri_validator), PlainSerializer(url_serializer)
]
"""
Absolute uniform resource identifier
"""

URIReference = Annotated[
    ParseResult, PlainValidator(generic_uri_validator), PlainSerializer(url_serializer)
]
"""
Uniform resource identifier reference
"""

DateTime = datetime
"""
Date and time expression using the Gregorian Calendar
"""


class SpecVersion(str, Enum):
    """
    The version of the CloudEvents specification which an event uses.
    This enables the interpretation of the context.

    Currently, this attribute will only have the 'major' and 'minor' version numbers
    included in it. This allows for 'patch' changes to the specification to be made
    without changing this property's value in the serialization.
    """

    # v0_3 = "0.3"
    v1_0 = "1.0"



# ==============================================================================
#  Copyright (c) 2024 Federico Busetti                                         =
#  <729029+febus982@users.noreply.github.com>                                  =
#                                                                              =
#  Permission is hereby granted, free of charge, to any person obtaining a     =
#  copy of this software and associated documentation files (the "Software"),  =
#  to deal in the Software without restriction, including without limitation   =
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    =
#  and/or sell copies of the Software, and to permit persons to whom the       =
#  Software is furnished to do so, subject to the following conditions:        =
#                                                                              =
#  The above copyright notice and this permission notice shall be included in  =
#  all copies or substantial portions of the Software.                         =
#                                                                              =
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  =
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    =
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     =
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  =
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     =
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         =
#  DEALINGS IN THE SOFTWARE.                                                   =
# ==============================================================================
import base64
import datetime
from typing import Annotated, Any, Optional, Union

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    model_serializer,
    model_validator,
)
from pydantic.fields import FieldInfo
from pydantic_core.core_schema import ValidationInfo
from ulid import ULID

from .fields.metadata import (
    FieldData,
    FieldDataContentType,
    FieldDataSchema,
    FieldId,
    FieldSource,
    FieldSpecVersion,
    FieldSubject,
    FieldTime,
    FieldType,
)
from .fields.types import URI, Binary, DateTime, SpecVersion, String, URIReference

DEFAULT_SPECVERSION = SpecVersion.v1_0


_binary_field_metadata = FieldInfo.from_annotation(Binary).metadata


class CloudEvent(BaseModel):  # type: ignore
    """
    A Python-friendly CloudEvent representation backed by Pydantic-modeled fields.
    """

    @classmethod
    def event_factory(
        cls,
        id: Optional[str] = None,
        specversion: Optional[SpecVersion] = None,
        time: Optional[Union[datetime.datetime, str]] = None,
        **kwargs,
    ) -> "CloudEvent":
        """
        Builds a new CloudEvent using sensible defaults.

        :param id: The event id, defaults to a ULID
        :type id: typing.Optional[str]
        :param specversion: The specversion of the event, defaults to 1.0
        :type specversion: typing.Optional[SpecVersion]
        :param time: The time the event occurred, defaults to now
        :type time: typing.Optional[Union[datetime.datetime, str]]
        :param kwargs: Other kwargs forwarded directly to the CloudEvent model.
        :return: A new CloudEvent model
        :rtype: CloudEvent
        """
        return cls(
            id=id or str(ULID()),
            specversion=specversion or DEFAULT_SPECVERSION,
            time=time or datetime.datetime.now(datetime.timezone.utc),
            **kwargs,
        )

    data: Annotated[Any, Field(default=None), FieldData]

    # Mandatory fields
    source: Annotated[URIReference, FieldSource]
    id: Annotated[String, FieldId]
    type: Annotated[String, FieldType]
    specversion: Annotated[SpecVersion, FieldSpecVersion]

    # Optional fields
    time: Annotated[Optional[DateTime], Field(default=None), FieldTime]
    subject: Annotated[Optional[String], Field(default=None), FieldSubject]
    datacontenttype: Annotated[
        Optional[String], Field(default=None), FieldDataContentType
    ]
    dataschema: Annotated[Optional[URI], Field(default=None), FieldDataSchema]

    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra={
            "example": {
                "specversion": "1.0",
                "type": "com.github.pull_request.opened",
                "source": "https://github.com/cloudevents/spec/pull",
                "subject": "123",
                "id": "A234-1234-1234",
                "time": "2018-04-05T17:31:00Z",
                "comexampleextension1": "value",
                "comexampleothervalue": 5,
                "datacontenttype": "text/xml",
                "data": '<much wow="xml"/>',
            }
        },
    )

    """
    Having the JSON functionality here is a violation of the Single Responsibility
    Principle, however we want to get advantage of improved pydantic JSON performances.
    Using `orjson` could solve this, perhaps it could be a future improvement.
    """

    # Typing for return value here breaks `.model_json_schema(mode='serialization')`
    @model_serializer(when_used="json")
    def base64_json_serializer(self):
        """Takes care of handling binary data serialization into `data_base64`
        attribute.

        :param self: CloudEvent.

        :return: Event serialized as a standard CloudEvent dict with binary
                 data handled.
        """
        model_dict = self.model_dump()  # type: ignore
        if _binary_field_metadata == self.model_fields["data"].metadata:
            model_dict["data_base64"] = model_dict["data"]
            del model_dict["data"]
        elif isinstance(model_dict["data"], (bytes, bytearray, memoryview)):
            model_dict["data_base64"] = base64.b64encode(model_dict["data"])
            del model_dict["data"]

        return model_dict

    @model_validator(mode="before")
    @classmethod
    def base64_json_validator(cls, data: dict, info: ValidationInfo) -> Any:
        """Takes care of handling binary data deserialization from `data_base64`
        attribute.

        :param data: Input data for validation
        :param info: Pydantic validation context
        :return: input data, after handling data_base64
        """
        if info.mode == "json" and isinstance(data, dict) and data.get("data_base64"):
            data["data"] = base64.b64decode(data["data_base64"])
            del data["data_base64"]
        return data
